unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ImgList, ComCtrls, StdCtrls, ExtCtrls, IdHTTP, IdIOHandlerStack,
  IdIntercept, IdGlobal, Math;

const
  iProgressing = 0;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    Btn_GetFile: TButton;
    Label_URLHint: TLabel;
    Ed_URL: TEdit;
    GroupBox1: TGroupBox;
    ListView_DownloadProgress: TListView;
    ImageList1: TImageList;
    SaveDialog1: TSaveDialog;
    procedure Btn_GetFileClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure ListView_DownloadProgressDeletion(Sender: TObject;
      Item: TListItem);
  private
    { Private declarations }
  public
    { Public declarations }
    FFormClosing : boolean;
  end;

  TSingleDownloadTask = class;

  TDownloadRefreshThread = class(TThread)
  private
    FOwner : TSingleDownloadTask;
  protected
    procedure Execute; override;
  public
    constructor Create(Owner : TSingleDownloadTask; CanSuspended : boolean);
    destructor Destroy; override;
  end;

  TSingleDownloadTask = class(TObject)
  private
    FMyItem : TListItem;
    FFileURL : WideString;
    FHTTPClient : TIdHTTP;
    FIOHandlerStack : TIdIOHandlerStack;
    FConnectionIntercept : TIdConnectionIntercept;
    FFileSize: Cardinal;
    FCurrentSize: Cardinal;
    FFilename: string;
    FMyThread : TDownloadRefreshThread;

    procedure SetCurrentSize(const Value: Cardinal);
    procedure SetFileSize(const Value: Cardinal);
    procedure SetFilename(const Value: string);
  public
    property CurrentSize : Cardinal read FCurrentSize write SetCurrentSize;
    property FileSize : Cardinal read FFileSize write SetFileSize;
    property Filename : string read FFilename write SetFilename;

    constructor Create(MyItem : TListItem; DownloadURL : String);
    destructor Destroy; override;

    procedure GetData;
    procedure ReceiveData(ASender: TIdConnectionIntercept;
               var ABuffer: TIdBytes);
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}


{$REGION ' TDownloadRefershThread class Implementation'}
  { TDownloadRefreshThread }
  
  constructor TDownloadRefreshThread.Create(Owner: TSingleDownloadTask;
    CanSuspended: boolean);
  begin
     inherited Create(CanSuspended);
  
     Self.FOwner := Owner;
     SElf.FreeOnTerminate := True;
  end;
  
  destructor TDownloadRefreshThread.Destroy;
  begin
    if Assigned(Self.FOwner) then begin
       Self.FOwner.FConnectionIntercept.OnReceive := nil;
       
       if Self.FOwner.FHTTPClient.Connected then
          Self.FOwner.FHTTPClient.Disconnect;
       Self.FOwner := nil;
    end;
  
    inherited;
  end;
  
  procedure TDownloadRefreshThread.Execute;
  begin
    try
       if not Form1.FFormClosing then begin
          self.FOwner.GetData;
       end else self.Terminate;
    finally
       Self.Terminate;
    end;
  end;
  
{$ENDREGION}


procedure TForm1.Btn_GetFileClick(Sender: TObject);
var
   TaskItem : TListItem;
   Task : TSingleDownloadTask;
begin
   if SElf.Ed_URL.Text <> '' then begin
      if MessageDlg('ЦۦT{zJ URL OӰjɮ, M|ܧִN], zNݤĪGF.' + #13 + #10 +
                    'M, oӽdҤ]|zT{ URL O_T, check  code Цۤv[J, n~?', mtInformation, [mbYes, mbNo], 0) = mrYes then begin
         if Self.SaveDialog1.Execute then begin
            TaskItem := ListView_DownloadProgress.Items.Add;
            TaskItem.Caption := self.Ed_URL.Text;
            TaskItem.SubItems.Add(Self.SaveDialog1.FileName);
            TaskItem.SubItems.Add('0'); // FileSize;
            TaskItem.SubItems.Add('0 Byte');

            Task := TSingleDownloadTask.Create(TaskItem, Self.Ed_URL.Text);
            Task.Filename := Self.SaveDialog1.FileName;
            Task.FMyThread.Resume;

            Form1.Btn_GetFile.Enabled := False;
         end;
      end;
   end else begin
       MessageDlg('zSJ URL, !', mtWarning, [mbOK], 0);
   end;
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
   FFormClosing := True;
end;

procedure TForm1.ListView_DownloadProgressDeletion(Sender: TObject;
  Item: TListItem);
var
   SingleThread : TSingleDownloadTask;
begin
   if Item.Data <> nil then begin
      SingleThread := TSingleDownloadTask(Item.Data);

      SingleThread.FMyThread.Terminate;

     //Item.Data := nil; Hno˶? Ϊ, ]b Thread  Free F.
   end;
end;





{ TSingleDownloadTask }

constructor TSingleDownloadTask.Create(MyItem: TListItem;
  DownloadURL: String);
begin
  inherited Create;

  Self.FFileURL := DownloadURL;
  FMyItem := MyItem;

  Self.FIOHandlerStack := TIdIOHandlerStack.Create;
  Self.FConnectionIntercept := TIdConnectionIntercept.Create;
  SElf.FHTTPClient := TIdHTTP.Create(nil);

  Self.FHTTPClient.IOHandler := Self.FIOHandlerStack;
  Self.FIOHandlerStack.Intercept := Self.FConnectionIntercept;
  Self.FConnectionIntercept.OnReceive := Self.ReceiveData;

  Self.FMyThread := TDownloadRefreshThread.Create(Self, True);
end;

destructor TSingleDownloadTask.Destroy;
begin
Self.FConnectionIntercept.OnReceive := nil;
  
  if not SElf.FMyThread.terminated then begin
     Self.FMyThread.Terminate;
  end;

  if Assigned(Self.FMyItem) then
     Self.FMyItem.Data := nil;

  Self.FMyItem := nil;

  if Self.FHTTPClient.Connected then
     Self.FHTTPClient.Disconnect(True);
  Self.FHTTPClient.Free;

  inherited;
end;

procedure TSingleDownloadTask.GetData;
var
   FileStrm : TFileStream;
begin
   FileStrm := nil;
   if FileExists(Self.FFilename) then begin
      try
         FileStrm := TFileStream.Create(SElf.FFilename, fmOpenReadWrite);
         FileStrm.Position := 0;
      except
         Self.FMyThread.Terminate;
         if Assigned(FileStrm) then
            FileStrm.Free;
      end;
   end else begin
       FileStrm := TFileStream.Create(SElf.FFilename, fmCreate);
   end;

   try
      if Assigned(FileStrm) then begin
         Self.FHTTPClient.Head(Self.FFileURL);
         ShowMessage(Self.FHTTPClient.Response.RawHeaders.Text);

         Self.FMyItem.SubItems.Strings[1] := Self.FHTTPClient.Response.RawHeaders.Values['Content-Length'] + ' Bytes';
         FFileSize := StrToInt(Self.FHTTPClient.Response.RawHeaders.Values['Content-Length']);
         Self.CurrentSize := 0;
         Self.FHTTPClient.Get(Self.FFileURL, FileStrm);
      end;
   finally
      if Assigned(FileStrm) then
         FileStrm.Free;
   end;
end;

procedure TSingleDownloadTask.ReceiveData(ASender: TIdConnectionIntercept;
  var ABuffer: TIdBytes);
begin
   CurrentSize := CurrentSize + Cardinal(Length(ABuffer));

   if not Form1.FFormClosing then begin
     if CurrentSize >= FFileSize then begin
        if not Form1.FFormClosing then begin
           if self.FMyItem.ListView.Showing then
              Self.FMyItem.ListView.Invalidate;
        end;
    end else begin
        if not Form1.FFormClosing then begin
           if Self.FMyItem.ImageIndex <> iProgressing then begin
              Self.FMyItem.ImageIndex := iProgressing;

              if self.FMyItem.ListView.Showing then
                 Self.FMyItem.ListView.Invalidate;
           end;
        end;
    end;
   end;

  Application.ProcessMessages;
end;

procedure TSingleDownloadTask.SetCurrentSize(const Value: Cardinal);
begin
  FCurrentSize := Value;

  if not Form1.FFormClosing then begin
     if Self.FFileSize > 0 then begin
        if Self.FCurrentSize <= Self.FFileSize then
           Self.FMyItem.SubItems.strings[2] := IntToStr(Value) + ' Bytes'
        else begin
             MessageDlg('U', mtInformation, [mbOK], 0);
             Self.FMyItem.ImageIndex := 1;
             Form1.Btn_GetFile.Enabled := True;
        end;
     end;
  end;
end;

procedure TSingleDownloadTask.SetFilename(const Value: string);
begin
  FFilename := Value;
end;

procedure TSingleDownloadTask.SetFileSize(const Value: Cardinal);
begin
  FFileSize := Value;
end;

end.
